/*************************************************************************************************/
/*                       Fragment Shader for Implicity Surface Ray Tracing                       */	
/*************************************************************************************************/

struct SCamera
{
	vec3 Position;
	
	vec3 Side;
	
	vec3 Up;
	
	vec3 View;
	
	vec2 Scale;
};

//-------------------------------------------------------------------------------------------------

struct SRay
{
	vec3 Origin;
	
	vec3 Direction;
};

//-------------------------------------------------------------------------------------------------

struct SJuliaFractal
{	
	vec4 Center;

	int Maximum;
	
	float Epsilon;
	
	float Delta;

	float EscapeThreshold;
};

///////////////////////////////////////////////////////////////////////////////////////////////////
// Shader Interface

uniform SCamera Camera;

uniform SJuliaFractal JuliaFractal;

///////////////////////////////////////////////////////////////////////////////////////////////////
// Shader Constants

const int NormalIterations = 10;

const float BoundingRadius = 3.0;

//-------------------------------------------------------------------------------------------------

const vec3 Zero = vec3 ( 0.0 );

const vec3 Unit = vec3 ( 1.0 );

//-------------------------------------------------------------------------------------------------

const vec3 AxisX = vec3 ( 1.0, 0.0, 0.0 );

const vec3 AxisY = vec3 ( 0.0, 1.0, 0.0 );

const vec3 AxisZ = vec3 ( 0.0, 0.0, 1.0 );

///////////////////////////////////////////////////////////////////////////////////////////////////
// Math Functions

vec4 QuaternionMult ( vec4 left, vec4 right )
{
   vec4 r;

   r.x = left.x * right.x - dot ( left.yzw, right.yzw );
   
   r.yzw = left.x * right.yzw + right.x * left.yzw + cross ( left.yzw, right.yzw );

   return r;
}

vec4 QuaternionSquare ( vec4 quaternion )
{
   vec4 r;

   r.x = quaternion.x * quaternion.x - dot ( quaternion.yzw, quaternion.yzw );
   
   r.yzw = 2.0 * quaternion.x * quaternion.yzw;

   return r;
}

///////////////////////////////////////////////////////////////////////////////////////////////////
// Intersection Functions

bool IntersectSphere ( SRay ray, float sqrradius, out float start )
{
	float A = dot ( ray.Direction, ray.Direction );

	float B = dot ( ray.Direction, ray.Origin );

	float C = dot ( ray.Origin, ray.Origin ) - sqrradius;

	float D = B * B - A * C;
	
	if ( D > 0.0 )
	{
		D = sqrt ( D );

		start = max ( 0.0, ( -B - D ) / A );

		return D > B;
	}

	return false;
}

void Iterate ( inout vec4 quaternion, inout vec4 derivative, SJuliaFractal fractal )
{
	for ( int i = 0; i < fractal.Maximum; i++ )
	{
		derivative = 2.0 * QuaternionMult ( quaternion, derivative );
	
		quaternion = QuaternionSquare ( quaternion ) + fractal.Center;
		
		if ( dot ( quaternion, quaternion ) > fractal.EscapeThreshold )
		{
			break;
		}
	}
}

float Intersect ( inout SRay ray, SJuliaFractal fractal )
{
	float distance = 0.0;
	
	while ( true )
	{
		vec4 quaternion = vec4 ( ray.Origin, 0.0 );
		
		vec4 derivative = vec4 ( 1.0, 0.0, 0.0, 0.0 );

		Iterate ( quaternion, derivative, fractal );

		float norm = length ( quaternion );
		
		distance = 0.5 * norm * log ( norm ) / length ( derivative );
		
		ray.Origin += ray.Direction * distance;
		
		if ( distance < fractal.Epsilon || dot ( ray.Origin, ray.Origin ) > BoundingRadius )
		{
			break;
		}
	}
	
	return distance;
}

vec3 NormEstimate ( vec3 point, SJuliaFractal fractal )
{
   vec4 quaternion = vec4 ( point, 0.0 );
   
   vec4 xLeft = quaternion - vec4 ( fractal.Delta, 0.0, 0.0, 0.0 );
   vec4 xRight = quaternion + vec4 ( fractal.Delta, 0.0, 0.0, 0.0 );
   
   vec4 yLeft = quaternion - vec4 ( 0.0, fractal.Delta, 0.0, 0.0 );
   vec4 yRight = quaternion + vec4 ( 0.0, fractal.Delta, 0.0, 0.0 );
   
   vec4 zLeft = quaternion - vec4 ( 0.0, 0.0, fractal.Delta, 0.0 );
   vec4 zRight = quaternion + vec4 ( 0.0, 0.0, fractal.Delta, 0.0 );

   for ( int counter = 0; counter < NormalIterations; counter++ )
   {
      xLeft = QuaternionSquare ( xLeft ) + fractal.Center;
      xRight = QuaternionSquare ( xRight ) + fractal.Center;
      
      yLeft = QuaternionSquare ( yLeft ) + fractal.Center;
      yRight = QuaternionSquare ( yRight ) + fractal.Center;
      
      zLeft = QuaternionSquare ( zLeft ) + fractal.Center;
      zRight = QuaternionSquare ( zRight ) + fractal.Center;
   }

   vec3 gradient = vec3 ( length ( xRight ) - length ( xLeft ),
                          length ( yRight ) - length ( yLeft ),
                          length ( zRight ) - length ( zLeft ) );
                        
   return normalize ( gradient );
}

///////////////////////////////////////////////////////////////////////////////////////////////////
// Support Functions

SRay GenerateRay ( SCamera camera )
{
	#if defined ( RENDER_TRUE_STEREO )

	float x = gl_TexCoord [0].x + mix ( -1.0, 1.0, float ( gl_TexCoord [0].x < 0.0 ) );

	vec2 coords = vec2 ( x, gl_TexCoord [0].y ) * camera.Scale;

	#else

	vec2 coords = gl_TexCoord [0].xy * camera.Scale;

	#endif
	
	vec3 direction = camera.View - camera.Side * coords.x + camera.Up * coords.y;
   
	return SRay ( camera.Position, normalize ( direction ) );
}

//-------------------------------------------------------------------------------------------------

#define K_A 0.2

#define K_D 0.8

#define K_S 0.8

#define P 64.0

vec3 Phong ( vec3 point, vec3 normal, vec3 color )
{
	vec3 light = normalize ( Camera.Position - point );
   
	float diffuse = abs ( dot ( light, normal ) );

	vec3 reflect = reflect ( -light, normal );

	float specular = pow ( abs ( dot ( reflect, light ) ), P );

	return K_A * Unit + diffuse * ( K_D * color + K_S * specular * Unit );
}

///////////////////////////////////////////////////////////////////////////////////////////////////
// Shader Entry Point

vec3 Raytrace ( SRay ray )
{
	vec3 result = Zero;

	float time = 0.0;
	
	if ( IntersectSphere ( ray, BoundingRadius, time ) )
	{
		ray.Origin += ray.Direction * time;
			
		time = Intersect ( ray, JuliaFractal );
			
		if ( time < JuliaFractal.Epsilon )
		{
			vec3 normal = NormEstimate ( ray.Origin, JuliaFractal );
				
			result += Phong ( ray.Origin, normal, abs ( ray.Origin ) );
		}
	}
	
	return result;
}

//-------------------------------------------------------------------------------------------------

#if defined ( TRUE_ANAGLYPH )

const mat3 MatrixL = mat3 ( 0.299, 0.000, 0.000,
					        0.587, 0.000, 0.000,
					        0.114, 0.000, 0.000 );

const mat3 MatrixR = mat3 ( 0.000, 0.000, 0.299,
					        0.000, 0.000, 0.587,
					        0.000, 0.000, 0.114 );

#elif defined ( COLOR_ANAGLYPH )

const mat3 MatrixL = mat3 ( 1.000, 0.000, 0.000,
					        0.000, 0.000, 0.000,
					        0.000, 0.000, 0.000 );

const mat3 MatrixR = mat3 ( 0.000, 0.000, 0.000,
					        0.000, 1.000, 0.000,
					        0.000, 0.000, 1.000 );

#elif defined ( HALF_COLOR_ANAGLYPH )

const mat3 MatrixL = mat3 ( 0.299, 0.000, 0.000,
					        0.587, 0.000, 0.000,
					        0.114, 0.000, 0.000 );

const mat3 MatrixR = mat3 ( 0.000, 0.000, 0.000,
					        0.000, 1.000, 0.000,
					        0.000, 0.000, 1.000 );

#elif defined ( GRAY_ANAGLYPH )

const mat3 MatrixL = mat3 ( 0.299, 0.000, 0.000,
					        0.587, 0.000, 0.000,
					        0.114, 0.000, 0.000 );

const mat3 MatrixR = mat3 ( 0.000, 0.299, 0.299,
					        0.000, 0.587, 0.587,
					        0.000, 0.114, 0.114 );

#elif defined ( OPTIMIZED_ANAGLYPH )

const mat3 MatrixL = mat3 ( 0.000, 0.000, 0.000,
					        0.700, 0.000, 0.000,
					        0.300, 0.000, 0.000 );

const mat3 MatrixR = mat3 ( 0.000, 0.000, 0.000,
					        0.000, 1.000, 0.000,
					        0.000, 0.000, 1.000 );

#endif

//-------------------------------------------------------------------------------------------------

void main ( void )
{
	vec3 result = Zero;

	//=============================================================================================

	#if defined ( RENDER_MONO )

	SRay ray = GenerateRay ( Camera );

	result = Raytrace ( ray );

	//=============================================================================================

	#elif defined ( RENDER_ANAGLYPH_FULL )

	SCamera camera = Camera;

	camera.Position = Camera.Position - Camera.Side * DELTA;

	SRay ray = GenerateRay ( camera );

	vec3 right = Raytrace ( ray );

	camera.Position = Camera.Position + Camera.Side * DELTA;

	ray = GenerateRay ( camera );

	vec3 left = Raytrace ( ray );

	result = MatrixL * left + MatrixR * right;

	#if defined ( OPTIMIZED_ANAGLYPH )

	result.r = pow ( result.r, 1.0 / 1.5) ;

	#endif

	//=============================================================================================

	#elif defined ( RENDER_ANAGLYPH_INTERLACED )

	SCamera camera = Camera;
	
	float line = mod ( floor ( gl_FragCoord.y ), 2.0 );
	
	camera.Position += Camera.Side * mix ( -DELTA, DELTA, line );
	
	SRay ray = GenerateRay ( camera );
	
	result = Raytrace ( ray );
	
	result = mix ( MatrixR * result, MatrixL * result, line );

	#if defined ( OPTIMIZED_ANAGLYPH )

	result.r = pow ( result.r, 1.0 / 1.5) ;

	#endif

	//=============================================================================================

	#elif defined ( RENDER_TRUE_STEREO )

	SCamera camera = Camera;

	float side = float ( gl_TexCoord [0].x < 0.0 );
	
	camera.Position += -Camera.Side * mix ( -DELTA, DELTA, side );
	
	SRay ray = GenerateRay ( camera );
	
	result = Raytrace ( ray );

	#endif

	//=============================================================================================
	
	gl_FragColor = vec4 ( result, 1.0 );
}